/*
	d80211 Marvell Libertas PCI driver

	Copyright (c) 2007 Andrea Merello <andreamrl@tiscali.it>
	
	Based on existent mrv8k driver
	Copyright (c) 2005 Luc Saillard <luc@saillard.org>
	
	Based on BSD malo driver (no code has been copyed. Programming
	methods, magic values, and card specific programming details
	has been looked at)

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; see the file COPYING.  If not, write to
	the Free Software Foundation, Inc., 51 Franklin Steet, Fifth Floor,
	Boston, MA 02110-1301, USA.

*/
#define DRV_COPYRIGHT 			"Andrea Merello <andreamrl *at* tiscali.it>"
#define DRV_DESCRIPTION 			"d80211 Marvell Libertas PCI driver"
#define DRV_VERSION 				"0.1"
#define DRV_NAME					"libertaspci"
#define FW_FIRMWARE_FILENAME	"mrv8k-f.fw"
#define FW_BOOT_FILENAME			"mrv8k-b.fw"

#define MRV_CMD_PKT_SIZE						512
#define MRV_CMD_MAGIC							0xAA55AA55
#define MRV_FW_SIGNATURE					0xf0f1f2f4

/* it is intentional we add sizeof(mrv_cmd_hdr) to all commands, but we don't add the header len
 * when we send the MRV_CMD_FW. Bootloader don't want the header len, while firmware seems to 
 * need it 
 */
#define MRV_CMD_FW						0x001
#define MRV_CMD_GET_CONFIG				0x003
#define MRV_CMD_GET_CONFIG_SZ 			(sizeof(struct mrv_cmd_get_config)+sizeof(struct mrv_cmd_hdr))
#define MRV_CMD_SET_RADIO				0x01c
#define MRV_CMD_SET_RADIO_SZ			(sizeof(struct mrv_cmd_set_radio)+ sizeof(struct mrv_cmd_hdr))
#define MRV_PREAMBLE_AUTO				5
#define MRV_PREAMBLE_SHORT				1
#define MRV_PREAMBLE_LONG				3
#define MRV_CMD_SET_TXPOWER			0x01e
#define MRV_CMD_SET_TXPOWER_SZ		(sizeof(struct mrv_cmd_set_txpower)+ sizeof(struct mrv_cmd_hdr))
#define MRV_CMD_SET_ANTENNA			0x020
#define MRV_CMD_SET_ANTENNA_SZ			(sizeof(struct mrv_cmd_set_antenna)+ sizeof(struct mrv_cmd_hdr))
#define MRV_ANTENNA_RXA					1
#define MRV_ANTENNA_TXA					2
#define MRV_ANTENNA_RXB					4
#define MRV_ANTENNA_TXB					8
#define MRV_CMD_SET_CHANNEL 			0x10a /*0x1d for usb*/
#define MRV_CMD_SET_CHANNEL_SZ			(sizeof(struct mrv_cmd_set_channel)+ sizeof(struct mrv_cmd_hdr))
#define MRV_CMD_SET_RATES				0x110 /*0x22 0x76 0x7f about rates USB*/
#define MRV_CMD_SET_RATES_SZ			(sizeof(struct mrv_cmd_set_rates)+ sizeof(struct mrv_cmd_hdr))
#define MRV_CMD_SET_RTS_THRESHOLD		0x113
#define MRV_CMD_SET_RTS_THRESHOLD_SZ	(sizeof(struct mrv_cmd_set_rts_threshold)+ sizeof(struct mrv_cmd_hdr))
#define MRV_CMD_SET_SLOT					0x114
#define MRV_CMD_SET_SLOT_SZ				(sizeof(struct mrv_cmd_set_slot)+ sizeof(struct mrv_cmd_hdr))
#define MRV_CMD_SET_MAC_ADDR			0x202
#define MRV_CMD_SET_MAC_ADDR_SZ		(sizeof(struct mrv_cmd_set_mac_addr)+ sizeof(struct mrv_cmd_hdr))

#define MRV_CMD_PAYLOAD(x) ( ((u8*)x) + sizeof(struct mrv_cmd_hdr))


#define MRV_INTA_MASK 0x1f
#define MRV_INTB_MASK 0x1f

#define MRV_REG_CMD_PTR			0xc10
#define MRV_REG_STATUS			0xc14
#define MRV_REG_GEN_ARM_INT		0xc18
#define MRV_REG_INTA_MASK		0xc34
#define MRV_REG_INTB_MASK		0xc3c
#define MRV_REG_ISR				0xc30
#define MRV_REG_RX_MODE			0xa300
#define MRV_REG_BSSID				0xa530
#define MRV_MEM_CMD_BOOT		0xbef8
#define MRV_HWMEM_CMD_BOOT	(0xc0000000 | MRV_MEM_CMD_BOOT)

#define PCI_DEVICE_ID_MARVELL_W8K (0x1FA6)
#define WL3563  0
#define WL138G 1
#define W8335 2
#define WLGENERIC 3

#define MRV_RX_DESC_NUM		128
#define MRV_RX_PAYLOAD_LEN	(2342 + 8)
#define MRV_MAX_RX_INT		128

#define MRV_TX_DESC_NUM		128

static const struct ieee80211_rate mrv_rates[] = {
	{ .rate = 10,
	  .val = 0 | (1<<4),
	  .flags = IEEE80211_RATE_CCK },
	{ .rate = 20,
	  .val = 1| (1<<4),
	  .flags = IEEE80211_RATE_CCK },
	{ .rate = 55,
	  .val = 2| (1<<4),
	  .flags = IEEE80211_RATE_CCK },
	{ .rate = 110,
	  .val = 3| (1<<4),
	  .flags = IEEE80211_RATE_CCK },
	{ .rate = 60,
	  .val = 4| (1<<4),
	  .flags = IEEE80211_RATE_OFDM },
	{ .rate = 90,
	  .val = 5| (1<<4),
	  .flags = IEEE80211_RATE_OFDM },
	{ .rate = 120,
	  .val = 6| (1<<4),
	  .flags = IEEE80211_RATE_OFDM },
	{ .rate = 180,
	  .val = 7| (1<<4),
	  .flags = IEEE80211_RATE_OFDM },
	{ .rate = 240,
	  .val = 8| (1<<4),
	  .flags = IEEE80211_RATE_OFDM },
	{ .rate = 360,
	  .val = 9| (1<<4),
	  .flags = IEEE80211_RATE_OFDM },
	{ .rate = 480,
	  .val = 10| (1<<4),
	  .flags = IEEE80211_RATE_OFDM },
	{ .rate = 540,
	  .val = 11| (1<<4),
	  .flags = IEEE80211_RATE_OFDM },
};

static const struct ieee80211_channel mrv_channels[] = {
	{ .chan = 1,
	  .freq = 2412},
	{ .chan = 2,
	  .freq = 2417},
	{ .chan = 3,
	  .freq = 2422},
	{ .chan = 4,
	  .freq = 2427},
	{ .chan = 5,
	  .freq = 2432},
	{ .chan = 6,
	  .freq = 2437},
	{ .chan = 7,
	  .freq = 2442},
	{ .chan = 8,
	  .freq = 2447},
	{ .chan = 9,
	  .freq = 2452},
	{ .chan = 10,
	  .freq = 2457},
	{ .chan = 11,
	  .freq = 2462},
	{ .chan = 12,
	  .freq = 2467},
	{ .chan = 13,
	  .freq = 2472},
	{ .chan = 14,
	  .freq = 2484}
};

struct mrv_rx_descriptor {
	u8	drvowner;
	u8	rssi;
	u8	status;			/* Need to be set to 1 */
	u8	chan;
	__le16	datalen;
	u8	rsvd0;
	u8	rate;
	__le32	dma_data;		/* Point to the dma buffer use to transfer the data */
	__le32	dma_next;		/* Point to the next dma buffer to transfer control data */
	__le16	qos;
	__le16	rsvd1;
} __attribute__((packed));

struct mrv_tx_descriptor {
	__le16	status;
	__le16	fwowner;
	u8	rate;
	u8	priority;
	__le16	qos;
	__le32	dma_addr;
	__le16	len;
	u8	dest_addr[6];
	__le32	dma_next;
	__le32	rsvd0;	
	__le32	rsvd1;
} __attribute__((packed));

struct mrv_priv {
	struct pci_dev *pdev;
	struct ieee80211_hw *dev;
	struct ieee80211_hw_mode modes[2];
	struct ieee80211_rate rates[ARRAY_SIZE(mrv_rates)];
	struct ieee80211_channel channels[ARRAY_SIZE(mrv_channels)];
	void *bar1;
	unsigned long bar1_phys;
	void *bar2;
	unsigned long bar2_phys;

	u8 *cmd_virt;
	dma_addr_t cmd_dma;
	__le32 *magic_virt;
	dma_addr_t magic_dma;

	int cmd_status;
	struct semaphore cmd_sem;

	struct tasklet_struct mrv_rx_tasklet;
	struct tasklet_struct mrv_tx_cb_tasklet;
	spinlock_t irq_th_lock;
	spinlock_t tx_lock;
	wait_queue_head_t wait_command_queue;

	int AntNum;
	int FwVer;
	int HwVer;
	int TxRingNum;
	int RegionCode; 
	int TxHwDma0;
	int TxHwDma1;
	int TxHwDma2;
	int TxHwDma3;
	int RxHwDmaR;
	int RxHwDmaW;

	struct sk_buff *rx_skb[MRV_RX_DESC_NUM];
	struct mrv_rx_descriptor *rx_ring_virt[MRV_RX_DESC_NUM];
	dma_addr_t rx_ring_dma[MRV_RX_DESC_NUM];
	int rx_idx;

	struct sk_buff *tx_skb[MRV_TX_DESC_NUM];
	struct mrv_tx_descriptor *tx_ring_virt[MRV_TX_DESC_NUM];
	dma_addr_t tx_ring_dma[MRV_TX_DESC_NUM];
	int tx_w_idx;
	int tx_r_idx;
	int tx_pending;

	u8 bssid[ETH_ALEN];
	int mode;
	int fw_load;
	int up;
};

static inline unsigned int mrv_reg_read8(struct mrv_priv *priv, unsigned int reg)
{
	return readb(priv->bar2 + reg);
}
static inline unsigned int mrv_reg_read16(struct mrv_priv *priv, unsigned int reg)
{
	return readw(priv->bar2 + reg);
}
static inline unsigned int mrv_reg_read32(struct mrv_priv *priv, unsigned int reg)
{
	return readl(priv->bar2 + reg);
}

static inline void mrv_reg_write8(struct mrv_priv *priv, unsigned int reg, __le32 value)
{
	writeb(value, (void *)(priv->bar2 + reg));
}
static inline void mrv_reg_write16(struct mrv_priv *priv, unsigned int reg, __le32 value)
{
	writew(value, (void *)(priv->bar2 + reg));
}
static inline void mrv_reg_write32(struct mrv_priv *priv, unsigned int reg, __le32 value)
{
	writel(value, (void *)(priv->bar2 + reg));
}

static inline unsigned int mrv_mem_read8(struct mrv_priv *priv, unsigned int reg)
{
	return readb(priv->bar1+ reg);
}
static inline unsigned int mrv_mem_read16(struct mrv_priv *priv, unsigned int reg)
{
	return readw(priv->bar1 + reg);
}
static inline unsigned int mrv_mem_read32(struct mrv_priv *priv, unsigned int reg)
{
	return readl(priv->bar1 + reg);
}

static inline void mrv_mem_write8(struct mrv_priv *priv, unsigned int reg, __le32 value)
{
	writeb(value, (void *)(priv->bar1 + reg));
}
static inline void mrv_mem_write16(struct mrv_priv *priv, unsigned int reg, __le32 value)
{
	writew(value, (void *)(priv->bar1 + reg));
}
static inline void mrv_mem_write32(struct mrv_priv *priv, unsigned int reg, __le32 value)
{
	writel(value, (void *)(priv->bar1 + reg));
}
static inline void mrv_reg_mb(struct mrv_priv *priv)
{
	mb();
	mrv_reg_read8(priv,0);
}

struct mrv_cmd_hdr { 
	__le16 type;
	__le16 len;
	__le16 sq;
	__le16 status;
} __attribute__ ((packed));

struct mrv_cmd_get_config {
	__le16	HwVer;
	__le16	TxRingNum;
	__le16	MCastAdrNum;
	u8	MacAdr[6];
	__le16	RegionCode;
	__le16	AntNum;
	__le32	FwVer;
	__le32	TxRing0;
	__le32	RxRingW;
	__le32	RxRingR;
	__le32	Magic;
	__le32	TxRing1;
	__le32	TxRing2;
	__le32	TxRing3;
} __attribute__((packed));

struct mrv_cmd_set_channel {
	__le16	flag;
	u8	ch;
//	__le16 rftype;
//	__le16 rsvd0;
//	u8	chlist[32];
} __attribute__((packed));

struct mrv_cmd_set_antenna {
	__le16	rtx;
	__le16	param;
} __attribute__((packed));

struct mrv_cmd_set_radio {
	__le16	flag;
	__le16	preamble;
	__le16	enable;
} __attribute__((packed));

struct mrv_cmd_set_rates {
	u8	fixedrate;
	u8	fixedindex;
	u8	rates[14];
} __attribute__((packed));

struct mrv_cmd_set_mac_addr {
	u8	addr[6];
} __attribute__((packed));

struct mrv_cmd_set_txpower {
	__le16	action;
	__le16	txpower;
	__le16	currentpowerlvl;
	__le16	rsvd;
	__le16	powers[8];
} __attribute__((packed));

struct mrv_cmd_set_rts_threshold{
	__le32	threshold;
} __attribute__((packed));


struct mrv_cmd_set_slot {
	__le16	action;
	u8	shortslot;
} __attribute__ ((packed));
